/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.util;

import java.awt.Polygon;

import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.ui.dnd.DropContext;
import org.fortiss.tooling.base.ui.layout.constants.DefaultLayoutConstants;

/**
 * Utility class for EditPart layout methods.
 * 
 * @author wandinger
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: C35B3098DFF4E3BD3F5B0AB20DC39B30
 */
// TODO: split class into sTg, rectangular and oval?
public class EditPartLayoutUtils {

	/** Returns the given value aligned to the given grid size. */
	public static int snapToGrid(double d, int gridSize) {
		return gridSize * (int) Math.round(d / gridSize);
	}

	/** Returns the given value aligned to the recommended grid size. */
	public static int snapToGrid(double d) {
		return snapToGrid(d, DefaultLayoutConstants.DEFAULT_GRID_SIZE);
	}

	/**
	 * Modifies the x and y coordinates of the given point to be on the grid of
	 * the given size.
	 */
	public static void snapToGrid(Point p, int gridSize) {
		p.x = snapToGrid(p.x, gridSize);
		p.y = snapToGrid(p.y, gridSize);
	}

	/** Modifies the x and y coordinates of the given point to be on the grid. */
	public static void snapToGrid(Point p) {
		p.x = snapToGrid(p.x, DefaultLayoutConstants.DEFAULT_GRID_SIZE);
		p.y = snapToGrid(p.y, DefaultLayoutConstants.DEFAULT_GRID_SIZE);
	}

	/**
	 * Returns polar coordinates of the point on the ellipse border having the
	 * given angle.
	 */
	public static Polar calculateEllipseBorderPolar(double angle,
			Dimension ellipseSize, int inset) {
		double a = ellipseSize.width / 2 - inset;
		double b = ellipseSize.height / 2 - inset;
		double epsilon = Math.sqrt(a * a - b * b) / a;
		double r = b
				/ Math.sqrt(1 - epsilon * epsilon * Math.cos(angle)
						* Math.cos(angle));
		return new Polar(angle, r);
	}

	/**
	 * Returns Cartesian coordinates of the point on the ellipse border having
	 * the given angle.
	 */
	public static Point calculateEllipseBorderPoint(double angle,
			Dimension ellipseSize, int inset) {
		Polar p = calculateEllipseBorderPolar(angle, ellipseSize, inset);
		return p.toCartesian().translate(ellipseSize.getScaled(.5));
	}

	/**
	 * Returns the {@link OffsetOrientation} of the point on the rectangle
	 * border having a given Point inside the rectangle.
	 */
	public static OffsetOrientation calculateRectangleBorderOrientation(
			Point source, Dimension rectangleSize, int inset) {
		int offset = 0;
		EOrientation orientation = EOrientation.NORTH;

		// The figure bounds
		Rectangle bounds = new Rectangle(0, 0, rectangleSize.width,
				rectangleSize.height);

		// Divide the rectangle into four diagonal sectors
		Polygon northSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.x, bounds.width }, new int[] { bounds.getCenter().y,
				bounds.y, bounds.y }, 3);
		Polygon eastSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.width, bounds.width }, new int[] { bounds.getCenter().y,
				bounds.y, bounds.height }, 3);
		Polygon southSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.width, bounds.x }, new int[] { bounds.getCenter().y,
				bounds.height, bounds.height }, 3);
		Polygon westSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.x, bounds.x }, new int[] { bounds.getCenter().y,
				bounds.height, bounds.y }, 3);

		// Determine in which sector we are and clamp the position accordingly
		if (northSector.contains(source.x, source.y)) {
			orientation = EOrientation.NORTH;
			offset = (int) (source.x * (1 - (2 * inset) / (double) bounds.width));
		}
		if (westSector.contains(source.x, source.y)) {
			orientation = EOrientation.WEST;
			offset = (int) (source.y * (1 - (2 * inset)
					/ (double) bounds.height));
		}
		if (southSector.contains(source.x, source.y)) {
			orientation = EOrientation.SOUTH;
			offset = (int) (source.x * (1 - (2 * inset) / (double) bounds.width));
		}
		if (eastSector.contains(source.x, source.y)) {
			orientation = EOrientation.EAST;
			offset = (int) (source.y * (1 - (2 * inset)
					/ (double) bounds.height));
		}
		return OffsetOrientationUtil.createOffsetOrientation(offset,
				orientation);
	}

	/**
	 * Ensure that the minimum size and aspect ratio of a oval shape are
	 * redeemed.
	 */
	public static Dimension clampOvalShapeSize(Dimension oldSize, int direction) {
		return clampOvalShapeSize(oldSize, direction,
				DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH,
				DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT,
				DefaultLayoutConstants.DEFAULT_SHAPE_ASPECT_RATIO);
	}

	/**
	 * Ensure that the minimum size and aspect ratio of a oval shape are
	 * redeemed.
	 */
	public static Dimension clampOvalShapeSize(Dimension oldSize,
			int direction, int minimumWidth, int minimumHeight,
			double aspectRatio) {
		Dimension newSize;

		// check for minimum size
		newSize = new Dimension(oldSize.width < minimumWidth ? minimumWidth
				: oldSize.width, oldSize.height < minimumHeight ? minimumHeight
				: oldSize.height);

		// ensure that the aspect ratio is redeemed
		switch (direction) {
		case PositionConstants.SOUTH_EAST:
			int aspectHeight = (int) (newSize.width / aspectRatio);
			int aspectWidth = (int) (newSize.height * aspectRatio);
			newSize.setSize(new Dimension(
					newSize.width > aspectWidth ? aspectWidth : newSize.width,
					newSize.height > aspectHeight ? aspectHeight
							: newSize.height));
			break;
		case PositionConstants.SOUTH:
			newSize.setSize(new Dimension((int) (newSize.height * aspectRatio),
					newSize.height));
			break;
		case PositionConstants.EAST:
			newSize.setSize(new Dimension(newSize.width,
					(int) (newSize.width / aspectRatio)));
			break;
		}

		return newSize;
	}

	/** Ensure that the minimum size of a rectangular shape is redeemed. */
	public static Dimension clampRectangularShapeSize(Dimension oldSize,
			int minimumWidth, int minimumHeight) {
		// check for minimum size
		Dimension newSize = new Dimension(
				oldSize.width < minimumWidth ? minimumWidth : oldSize.width,
				oldSize.height < minimumHeight ? minimumHeight : oldSize.height);
		return newSize;
	}

	/** Ensure that the minimum size of a rectangular shape is redeemed. */
	public static Dimension clampRectangularShapeSize(Dimension oldSize) {
		return clampRectangularShapeSize(oldSize,
				DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH,
				DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT);
	}

	/** Layouts the given connector within the surrounding ellipse. */
	public static void layoutConnectorInEllipse(ILayoutedModelElement layouted,
			ILayoutedModelElement container, Point location) {
		Point position = location.getCopy();
		org.fortiss.tooling.base.model.layout.Point np = LayoutDataUtils
				.getNodePosition(container);
		position.translate(-np.getX(), -np.getY());

		org.fortiss.tooling.base.model.layout.Dimension nd = LayoutDataUtils
				.getNodeSize(container);
		Dimension d = new Dimension(nd.getWidth(), nd.getHeight());
		position.translate(d.scale(-.5));

		// store the value in the model
		LayoutDataUtils.setConnectorAngle(layouted, new Polar(position).angle);
	}

	/** Layouts the given connector within the surrounding rectangle. */
	public static void layoutConnectorInRectangle(
			ILayoutedModelElement layouted, ILayoutedModelElement container,
			Point location) {
		Point position = location.getCopy();
		org.fortiss.tooling.base.model.layout.Point np = LayoutDataUtils
				.getNodePosition(container);
		position.translate(-np.getX(), -np.getY());
		// do border clamping
		org.fortiss.tooling.base.model.layout.Dimension nd = LayoutDataUtils
				.getNodeSize(container);
		Dimension d = new Dimension(nd.getWidth(), nd.getHeight());
		OffsetOrientation orient = EditPartLayoutUtils
				.calculateRectangleBorderOrientation(position, d,
						DefaultLayoutConstants.DEFAULT_SHAPE_INSETS);
		// snap to grid and store the value in the model
		orient.setOffset(EditPartLayoutUtils.snapToGrid(orient.getOffset()));
		LayoutDataUtils.setConnectorOffsetOrientation(layouted, orient);
	}

	/** Layouts the given layouted object using the position in the context. */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted,
			DropContext context, int positionOffset, boolean snapToGrid) {
		layoutNodeInRoot(layouted, context.getPosition(), positionOffset,
				snapToGrid);
	}

	/** Layouts the given layouted object using the position in the context. */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted,
			Point location, int positionOffset, boolean snapToGrid) {
		Point position = location.getCopy();
		position.translate(-positionOffset, -positionOffset);
		if (snapToGrid) {
			EditPartLayoutUtils.snapToGrid(position);
		}
		LayoutDataUtils.setNodePosition(layouted, position.x, position.y);
	}

	/**
	 * Class for storing polar coordinates.
	 * 
	 * @author wandinger
	 * @author hummel
	 * @author $Author$
	 * @version $Rev$
	 * @levd.rating YELLOW Rev: 519
	 */
	public static class Polar {

		/** The angle */
		public double angle;

		/** The distance */
		public double distance;

		/** The constructor */
		public Polar(double angle, double distance) {
			this.angle = angle;
			this.distance = distance;
		}

		/** Create new polar values from cartesian ones. */
		public Polar(Point pos) {
			angle = Math.atan2(pos.y, pos.x);
			distance = Math.sqrt(pos.x * pos.x + pos.y * pos.y);
		}

		/** Returns a cartesian representation. */
		public Point toCartesian() {
			return new Point(distance * Math.cos(angle), distance
					* Math.sin(angle));
		}
	}
}
