/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout.util;

import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.model.layout.Angle;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.LayoutFactory;
import org.fortiss.tooling.base.model.layout.Offset;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.base.ui.layout.constants.DefaultLayoutConstants;
import org.fortiss.tooling.base.ui.layout.constants.LayoutKeyConstants;

/**
 * Utility methods for {@link ILayoutedModelElement} and {@link ILayoutData}.
 * This class provides access to layout data using the default keys defined in
 * {@link LayoutKeyConstants}. Changing some layout data is done by replacing
 * the layout data object under a given key, which again results in a EMF
 * notification on the {@link ILayoutedModelElement}. Therefore a single change
 * notification listener on the layouted element can be used.
 * 
 * @author hoelzl
 * @author wandinger
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: F7FBEB4FD3CE4087069F5CEF38C76851
 */
public class LayoutDataUtils {

	/** Translates the given point by moving it accordingly. */
	public static Point translate(Point p, int dx, int dy) {
		return PointUtil.createPoint(p.getX() + dx, p.getY() + dy);
	}

	/** Expands the given dimension by resizing it accordingly. */
	public static Dimension expand(Dimension d, int dw, int dh) {
		return DimensionUtil.createDimension(d.getWidth() + dw, d.getHeight()
				+ dh);
	}

	/** Returns the bounds rectangle of the given layouted element. */
	public static Rectangle getNodeBounds(ILayoutedModelElement layouted) {
		Rectangle result = new Rectangle();
		Point p = getNodePosition(layouted);
		result.setLocation(p.getX(), p.getY());
		Dimension d = getNodeSize(layouted);
		result.setSize(d.getWidth(), d.getHeight());
		return result;
	}

	/** Sets the bounds rectangle of the given layouted node. */
	public static void setNodeBounds(ILayoutedModelElement layouted,
			Rectangle bounds) {
		setNodePosition(layouted, bounds.x, bounds.y);
		setNodeSize(layouted, bounds.width, bounds.height);
	}

	/** Returns the position of the given layouted node. */
	public static Point getNodePosition(ILayoutedModelElement layouted) {
		return PointUtil
				.getPosition(layouted, LayoutKeyConstants.NODE_POSITION);
	}

	/** Sets the layouted node's position. */
	// TODO: clean up should remove this
	public static void setNodePosition(ILayoutedModelElement layouted,
			Point position) {
		setNodePosition(layouted, position.getX(), position.getY());
	}

	/** Sets the layouted node's position. */
	public static void setNodePosition(ILayoutedModelElement layouted, int x,
			int y) {
		PointUtil.setPoint(layouted, LayoutKeyConstants.NODE_POSITION, x, y);
	}

	/**
	 * Returns the size of the given layouted node. If no dimension is stored in
	 * the given element a new dimension with the default size is returned.
	 */
	public static Dimension getNodeSize(ILayoutedModelElement layouted) {
		Dimension size = DimensionUtil.getDimension(layouted,
				LayoutKeyConstants.NODE_DIMENSION);

		if (size == null) {
			Dimension result = DimensionUtil.createDimension(
					DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH,
					DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT);
			result.setKey(LayoutKeyConstants.NODE_DIMENSION);
			return result;
		}
		return EcoreUtil.copy(size);
	}

	/** Sets the layouted node's size. */
	// TODO: cleanup should remove this method
	public static void setNodeSize(ILayoutedModelElement layouted,
			Dimension size) {
		setNodeSize(layouted, size.getWidth(), size.getHeight());
	}

	/** Sets the layouted node's size. */
	public static void setNodeSize(ILayoutedModelElement layouted, int w, int h) {
		DimensionUtil.setDimension(layouted, LayoutKeyConstants.NODE_DIMENSION,
				w, h);
	}

	/** Returns the layouted connector's position. */
	public static Point getConnectorPosition(ILayoutedModelElement layouted) {
		return PointUtil.getPosition(layouted,
				LayoutKeyConstants.CONNECTOR_POSITION);
	}

	/** Sets the layouted connector's position. */
	public static void setConnectorPosition(ILayoutedModelElement layouted,
			int x, int y) {
		PointUtil.setPoint(layouted, LayoutKeyConstants.CONNECTOR_POSITION, x,
				y);
	}

	/** Returns the layouted connector's orientation. */
	public static OffsetOrientation getConnectorOffsetOrientation(
			ILayoutedModelElement layouted) {
		OffsetOrientation oo = OffsetOrientationUtil.getOffsetOrientation(
				layouted, LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION);
		if (oo == null) {
			return OffsetOrientationUtil.createOffsetOrientation(0,
					EOrientation.NORTH);
		}
		return oo;
	}

	/** Sets the layouted connector's offset and orientation. */
	public static void setConnectorOffsetOrientation(
			ILayoutedModelElement layouted, OffsetOrientation orientation) {
		OffsetOrientationUtil.setOffsetOrientation(layouted,
				LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION,
				orientation.getOffset(), orientation.getOrientation());
	}

	/** Returns the layouted connector's offset. */
	public static Offset getConnectorOffset(ILayoutedModelElement layouted) {
		OffsetOrientation offset = OffsetOrientationUtil.getOffsetOrientation(
				layouted, LayoutKeyConstants.CONNECTOR_OFFSET);
		if (offset == null) {
			return OffsetOrientationUtil.createOffsetOrientation(0,
					EOrientation.NORTH);
		}
		return offset;
	}

	/** Sets the layouted connector's angle. */
	public static void setConnectorAngle(ILayoutedModelElement layouted,
			double angle) {
		AngleUtil.setAngle(layouted, LayoutKeyConstants.CONNECTOR_ANGLE, angle);
	}

	/** Returns the layouted connector's angle. */
	// TODO: used?
	public static Angle getConnectorAngle(ILayoutedModelElement layouted) {
		Angle angle = AngleUtil.getAngle(layouted,
				LayoutKeyConstants.CONNECTOR_ANGLE);
		if (angle == null) {
			return AngleUtil.createAngle(0.0);
		}
		return angle;
	}

	/** Returns the angle as double value (0 if not existing). */
	// TODO: used?
	public static double getConnectorAngleAsDouble(ILayoutedModelElement lobject) {
		return getConnectorAngle(lobject).getAngle();
	}

	/** Creates a new connection layout. */
	// TODO: used?
	public static void createConnectionLayout(ILayoutedModelElement lo) {
		Points points = PointsUtil.createPoints();
		points.setKey(LayoutKeyConstants.CONNECTION_POINTS);
		lo.getLayoutDataList().add(points);
	}

	/** Creates a new connector layout. */
	public static void createConnectorLayout(ILayoutedModelElement lo) {
		createNodeLayout(lo);

		Point cpos = LayoutFactory.eINSTANCE.createPoint();
		cpos.setKey(LayoutKeyConstants.CONNECTOR_POSITION);

		Dimension cdim = LayoutFactory.eINSTANCE.createDimension();
		cdim.setKey(LayoutKeyConstants.CONNECTOR_DIMENSION);

		lo.getLayoutDataList().add(cpos);
		lo.getLayoutDataList().add(cdim);
	}

	/** Creates a new node layout. */
	// TODO: public?
	public static void createNodeLayout(ILayoutedModelElement lobject) {
		PointUtil.setPoint(lobject, LayoutKeyConstants.NODE_POSITION, 0, 0);
		DimensionUtil.setDimension(lobject, LayoutKeyConstants.NODE_DIMENSION,
				DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH,
				DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT);
	}

	/**
	 * Return the absolute position for the given orientation using default
	 * insets.
	 */
	public static Point getAbsolute(OffsetOrientation oo, Dimension parentSize) {
		return getAbsolute(oo, parentSize,
				DefaultLayoutConstants.DEFAULT_SHAPE_INSETS);
	}

	/** Return the absolute position for a given offset and orientation. */
	// TODO: public?
	public static Point getAbsolute(OffsetOrientation oo, Dimension parentSize,
			int parentInsets) {
		Point pos = PointUtil.createPoint(0, 0);
		int offsetValue = oo.getOffset();
		switch (oo.getOrientation()) {
		case NORTH:
			pos.setX(offsetValue);
			pos.setY(0);
			break;
		case WEST:
			pos.setX(0);
			pos.setY(offsetValue);
			break;
		case SOUTH:
			pos.setX(offsetValue);
			pos.setY(parentSize.getHeight() - 2 * parentInsets);
			break;
		case EAST:
			pos.setX(parentSize.getWidth() - 2 * parentInsets);
			pos.setY(offsetValue);
			break;
		}

		return pos;
	}

	/** Returns the {@link Points} layout data object for the given connection. */
	public static Points getConnectionPoints(ILayoutedModelElement lobject) {
		return PointsUtil.getPoints(lobject,
				LayoutKeyConstants.CONNECTION_POINTS);
	}

	/** Sets the points layout data of the given connection. */
	private static void setConnectionPoints(ILayoutedModelElement lobject,
			Points points) {
		PointsUtil.setPoints(lobject, LayoutKeyConstants.CONNECTION_POINTS,
				points);
	}

	/** Adds the given x/y point to the layout data of the given element. */
	public static void addConnectionPoint(ILayoutedModelElement lobject, int x,
			int y) {
		Points pts = getConnectionPoints(lobject);
		Point np = PointUtil.createPoint(x, y);
		pts.getPointsList().add(np);
		setConnectionPoints(lobject, pts);
	}

	/** Changes the point of the layout data at the given index. */
	public static void changeConnectionPoint(ILayoutedModelElement lobject,
			int index, int x, int y) {
		Points pts = getConnectionPoints(lobject);
		Point np = PointUtil.createPoint(x, y);
		pts.getPointsList().set(index, np);
		setConnectionPoints(lobject, pts);
	}

	/** Changes all the point layout data of the given element. */
	public static void changeAllConnectionPoints(ILayoutedModelElement lobject,
			int dx, int dy) {
		Points pts = getConnectionPoints(lobject);
		Points npts = PointsUtil.createPoints();
		for (Point p : pts.getPointsList()) {
			Point np = PointUtil.createPoint(p.getX() + dx, p.getY() + dy);
			npts.getPointsList().add(np);
		}
		setConnectionPoints(lobject, npts);
	}

	/**
	 * Removes the point at the given index from the layout data of the given
	 * element.
	 */
	public static void removeConnectionPoint(ILayoutedModelElement lobject,
			int index) {
		Points pts = getConnectionPoints(lobject);
		pts.getPointsList().remove(index);
		setConnectionPoints(lobject, pts);
	}
}
